This is a test file
$(template_file.data_container[$(template_file.index)][bogus]) $(template_file.data_container[$(template_file.index)][foo]) $(template_file.data_container[$(template_file.index)][bar])
more stuff
